@@ -2,14 +2,15 @@ package Test::Mojo::Role::Selenium;
 use Mojo::Base -base;
 use Role::Tiny;
 
+use Carp 'croak';
 use File::Basename ();
 use File::Spec;
 use Mojo::Util qw(encode monkey_patch);
-use Selenium::Remote::WDKeys;
+use Selenium::Remote::WDKeys ();
 
 use constant DEBUG => $ENV{MOJO_SELENIUM_DEBUG} || 0;
 
-our $VERSION = '0.02';
+our $VERSION = '0.03';
 
 my $SCRIPT_NAME = File::Basename::basename($0);
 my $SCREENSHOT  = 1;
@@ -18,7 +19,7 @@ has driver => sub {
   my $self   = shift;
   my $args   = $self->driver_args;
   my $driver = $ENV{MOJO_SELENIUM_DRIVER} || $args->{driver_class} || 'Selenium::PhantomJS';
-  eval "require $driver;1" or die $@;
+  eval "require $driver;1" or croak "require $driver: $@";
   warn "[Selenium] Using $driver\n" if DEBUG;
   $driver = $driver->new(%$args, ua => $self->ua);
   $driver->debug_on if DEBUG > 1;
@@ -38,7 +39,7 @@ has _live_base => sub {
 
 has _live_server => sub {
   my $self   = shift;
-  my $app    = $self->app or die 'Cannot start server without $t->app(...) set';
+  my $app    = $self->app or croak 'Cannot start server without $t->app(...) set';
   my $server = Mojo::Server::Daemon->new(silent => DEBUG ? 0 : 1);
 
   Scalar::Util::weaken($self);
@@ -77,10 +78,10 @@ sub capture_screenshot {
 }
 
 sub click_ok {
-  my ($self, $selector, $desc) = @_;
+  my ($self, $selector) = @_;
   my $el = $selector ? $self->_proxy(find_element => $selector) : $self->driver->get_active_element;
   $el->click if $el;
-  return $self->_test('ok', $el, _desc($desc, "click on $selector"));
+  return $self->_test('ok', $el, _desc("click on $selector"));
 }
 
 sub current_url_is {
@@ -110,6 +111,9 @@ sub element_is_hidden {
   return $self->_test('ok', ($el && $el->is_hidden), _desc($desc, "element $selector is hidden"));
 }
 
+sub go_back    { $_[0]->_proxy('go_back');    $_[0] }
+sub go_forward { $_[0]->_proxy('go_forward'); $_[0] }
+
 sub live_element_count_is {
   my ($self, $selector, $count, $desc) = @_;
   my $els = $self->_proxy(find_elements => $selector);
@@ -131,14 +135,34 @@ sub live_element_exists_not {
 
 sub live_text_is {
   my ($self, $selector, $value, $desc) = @_;
-  return $self->_test('is', $self->_live_text($selector),
-    $value, _desc($desc, qq{exact match for selector "$selector"}));
+  return $self->_test(
+    'is', $self->_element_data(get_text => $selector),
+    $value, _desc($desc, qq{exact text for selector "$selector"})
+  );
 }
 
 sub live_text_like {
   my ($self, $selector, $regex, $desc) = @_;
-  return $self->_test('like', $self->_live_text($selector),
-    $regex, _desc($desc, qq{similar match for selector "$selector"}));
+  return $self->_test(
+    'like', $self->_element_data(get_text => $selector),
+    $regex, _desc($desc, qq{similar text for selector "$selector"})
+  );
+}
+
+sub live_value_is {
+  my ($self, $selector, $value, $desc) = @_;
+  return $self->_test(
+    'is', $self->_element_data(get_value => $selector),
+    $value, _desc($desc, qq{exact value for selector "$selector"})
+  );
+}
+
+sub live_value_like {
+  my ($self, $selector, $regex, $desc) = @_;
+  return $self->_test(
+    'like', $self->_element_data(get_value => $selector),
+    $regex, _desc($desc, qq{similar value for selector "$selector"})
+  );
 }
 
 sub navigate_ok {
@@ -169,10 +193,32 @@ around new => sub {
   return $self;
 };
 
+sub refresh { $_[0]->_proxy('refresh'); $_[0] }
+
 sub send_keys_ok {
   my ($self, $selector, $keys, $desc) = @_;
   my $el = $self->_proxy(find_element => $selector);
-  map { $el->send_keys(ref $_ ? KEYS->{$_} : $_) } ref $keys ? @$keys : ($keys) if $el;
+  my @keys;
+
+  for (ref $keys ? @$keys : ($keys)) {
+    my $key = ref $_ ? Selenium::Remote::WDKeys::KEYS()->{$$_} : $_;
+    croak "Invalid key '@{[ref $_ ? $$_ : $_]}'" unless defined $key;
+    push @keys, $key;
+  }
+
+  if ($el) {
+    eval {
+      for my $key (@keys) {
+        warn "[Selenium] send_keys $selector <- @{[Mojo::Util::url_escape($key)]}\n" if DEBUG;
+        $el->send_keys($key);
+      }
+      1;
+    } or do {
+      Test::More::diag($@);
+      $el = undef;
+    };
+  }
+
   return $self->_test('ok', $el, _desc($desc, "keys sent to $selector"));
 }
 
@@ -189,10 +235,28 @@ sub submit_ok {
   return $self->_test('ok', $el, _desc($desc, "click on $selector"));
 }
 
+sub wait_for_url {
+  my ($self, $regex, $desc) = @_;
+  my $driver = $self->driver;
+  my $guard  = 600;             # 60 seconds
+  my $ok;
+
+  while ($guard--) {
+    Time::HiRes::usleep(1e5);
+    warn "[Selenium] current_url=@{[$driver->get_current_url]}\n" if DEBUG > 2;
+    next unless $driver->get_current_url =~ /$regex/;
+    $ok = 1;
+    last;
+  }
+
+  return $self->_test('ok', $ok, _desc($desc, 'got expected url'));
+}
+
 sub window_size_is {
   my ($self, $exp, $desc) = @_;
   my $size = $self->driver->get_window_size;
-  $self->_test('is_deeply', [@$size{qw(width height)}],
+
+  return $self->_test('is_deeply', [@$size{qw(width height)}],
     $exp, _desc($desc, "window size is $exp->[0]x$exp->[1]"));
 }
 
@@ -217,10 +281,10 @@ sub _proxy {
   return $res;
 }
 
-sub _live_text {
-  my $self = shift;
+sub _element_data {
+  my ($self, $method) = (shift, shift);
   my $el = $self->_proxy(find_element => shift);
-  return $el ? $el->get_text : '';
+  return $el ? $el->$method : '';
 }
 
 sub _screenshot_name {
@@ -255,6 +319,34 @@ Test::Mojo::Role::Selenium - Test::Mojo in a real browser
 
 =head1 SYNOPSIS
 
+=head2 External app
+
+  use Mojo::Base -strict;
+  use Test::Mojo::WithRoles "Selenium";
+  use Test::More;
+
+  my $t = Test::Mojo::WithRoles->new;
+
+  $ENV{MOJO_SELENIUM_BASE_URL} = $ENV{TEST_BASE_URL} || '';
+  $ENV{MOJO_SELENIUM_DRIVER} ||= 'Selenium::Chrome';
+
+  plan skip_all => $@ || 'TEST_BASE_URL=http://mojolicious.org'
+    unless $ENV{TEST_BASE_URL} and eval { $t->driver };
+
+  $t->navigate_ok('/perldoc')
+    ->live_text_is('a[href="#GUIDES"]' => 'GUIDES');
+
+  $t->driver->execute_script(qq[document.querySelector("form").removeAttribute("target")]);
+  $t->element_is_displayed("input[name=q]")
+    ->send_keys_ok("input[name=q]", ["render", \"return"]);
+
+  $t->wait_for_url(qr{q=render})
+    ->live_value_is("input[name=search]", "render");
+
+  done_testing;
+
+=head2 Internal app
+
   use Mojo::Base -strict;
   use Test::Mojo::WithRoles "Selenium";
   use Test::More;
@@ -279,10 +371,10 @@ Test::Mojo::Role::Selenium - Test::Mojo in a real browser
     ->send_keys_ok("input[name=q]", "Mojo")
     ->capture_screenshot;
 
-  $t->submit_ok
+  $t->submit_ok("form")
     ->status_is(200)
     ->current_url_like(qr{q=Mojo})
-    ->value_is("input[name=q]", "Mojo");
+    ->live_element_exists("input[name=q][value=Mojo]");
 
   $t->click_ok("nav a.logo")->status_is(200);
 
@@ -386,25 +478,13 @@ Where screenshots are saved.
 
   $self = $self->active_element_is("input[name=username]");
 
-Test that the current active element on the page match the selector.
-
-=head2 button_down
-
-  $self = $self->button_down;
-
-See L<Selenium::Remote::Driver/button_down>.
-
-=head2 button_up
-
-  $self = $self->button_up;
-
-See L<Selenium::Remote::Driver/button_up>.
+Checks that the current active element on the page match the selector.
 
 =head2 capture_screenshot
 
   $self = $self->capture_screenshot;
   $self = $self->capture_screenshot("%t-page-x");
-  $self = $self->capture_screenshot("%0-%t-%n");
+  $self = $self->capture_screenshot("%0-%t-%n"); # default
 
 Capture screenshot to L</screenshot_directory> with filename specified by the
 input format. The format supports these special strings:
@@ -417,7 +497,7 @@ input format. The format supports these special strings:
 
 =head2 click_ok
 
-  $self = $self->click_ok("a", "left");
+  $self = $self->click_ok("a");
 
 Click on an element.
 
@@ -426,13 +506,14 @@ Click on an element.
   $self = $self->current_url_is("http://mojolicious.org/");
   $self = $self->current_url_is("/whatever");
 
-Test the current browser URL.
+Test the current browser URL against an absolute URL. A relative URL will be
+converted to an absolute URL, using L</MOJO_SELENIUM_BASE_URL>.
 
 =head2 current_url_like
 
   $self = $self->current_url_like(qr{/whatever});
 
-Test the current browser URL.
+Test the current browser URL against a regex.
 
 =head2 element_is_displayed
 
@@ -454,24 +535,42 @@ See L<Selenium::Remote::WebElement/is_hidden>.
 
   $self = $self->go_back;
 
+Equivalent to hitting the back button on the browser.
+
 See L<Selenium::Remote::Driver/go_back>.
 
 =head2 go_forward
 
   $self = $self->go_forward;
 
+Equivalent to hitting the forward button on the browser.
+
 See L<Selenium::Remote::Driver/go_forward>.
 
 =head2 live_element_count_is
 
+  $self = $self->live_element_count_is("a", 12);
+
+Checks that the selector finds the correct number of elements in the browser.
+
 See L<Test::Mojo/element_count_is>.
 
 =head2 live_element_exists
 
+  $self = $self->live_element_exists("div.content");
+
+Checks that the selector finds an element in the browser.
+
 See L<Test::Mojo/element_exists>.
 
 =head2 live_element_exists_not
 
+  $self = $self->live_element_exists_not("div.content");
+
+Checks that the selector does not find an element in the browser.
+
+  $self = $self->live_element_exists("div.foo");
+
 See L<Test::Mojo/element_exists_not>.
 
 =head2 live_text_is
@@ -488,6 +587,20 @@ browser matches the given string.
 Checks text content of the CSS selectors first matching HTML element in the
 browser matches the given regex.
 
+=head2 live_value_is
+
+  $self = $self->live_value_is("div.name", "Mojo");
+
+Checks value of the CSS selectors first matching HTML element in the browser
+matches the given string.
+
+=head2 live_value_like
+
+  $self = $self->live_value_like("div.name", qr{Mojo});
+
+Checks value of the CSS selectors first matching HTML element in the browser
+matches the given regex.
+
 =head2 navigate_ok
 
   $self = $self->navigate_ok("/");
@@ -495,16 +608,12 @@ browser matches the given regex.
 
 Open a browser window and go to the given location.
 
-=head2 maximize_window
-
-  $self = $self->maximize_window;
-
-See L<Selenium::Remote::Driver/maximize_window>.
-
 =head2 refresh
 
   $self = $self->refresh;
 
+Equivalent to hitting the refresh button on the browser.
+
 See L<Selenium::Remote::Driver/refresh>.
 
 =head2 send_keys_ok
@@ -515,13 +624,21 @@ See L<Selenium::Remote::Driver/refresh>.
 Used to sen keys to a given element. Scalar refs will be sent as
 L<Selenium::Remote::WDKeys> strings.
 
-=head2 set_selected_element_ok
+List of some of the special keys:
 
-  $self = $self->set_selected_element_ok("input[name=email]");
+=over 2
 
-Select and option, checkbox or radiobutton.
+=item * alt, control, shift
 
-See L<Selenium::Remote::WebElement/set_selected>
+=item * right_arrow, down_arrow, left_arrow, up_arrow
+
+=item * backspace, clear, delete, enter, return, escape, space, tab
+
+=item * f1, f2, ..., f12
+
+=item * command_meta, pause
+
+=back
 
 =head2 set_window_size
 
@@ -533,12 +650,18 @@ Set the browser window size.
 =head2 submit_ok
 
   $self = $self->submit_ok("form");
-  $self = $self->submit_ok;
 
 Submit a form, either by selector or the current active form.
 
 See L<Selenium::Remote::WebElement/submit>.
 
+=head2 wait_for_url
+
+  $self = $self->wait_for_url(qr{/whatever});
+
+Will sleep until the URL in the browser changes to something that matches the
+given regex.
+
 =head2 window_size_is
 
   $self = $self->window_size_is([$width, $height]);
